theme = [
    ('list-header', '', '', '', '', ''),
    ('list-header-column', 'underline', '', '', 'underline', ''),
    ('list-header-column-separator', '', '', '', '', ''),
    ('striped-table-row', 'white', 'dark gray', '', 'white', 'g27'),
    ('reveal focus', 'white', 'dark blue', 'standout', 'white', 'dark blue'),
    ('message status', '', '', '', '', ''),
    ('message error', 'white', 'dark red', 'standout', 'white', 'dark red'),
    ('status', 'dark blue', 'black', '', 'dark blue', 'black'),
    ('flash off', 'black', 'black', 'standout', 'black', 'black'),
    ('flash on', 'white', 'black', 'standout', 'white', 'black'),
    ('pop_up', 'white', 'black', '', 'white', 'black'),
    ('button action', 'white', 'light red', '', 'white', 'light red'),
    ('button cancel', 'black', 'light gray', '', 'black', 'light gray'),
]
